/*-
 * Copyright (c) 2003, Patrick MARIE <mycroft@virgaria.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <stdarg.h>

#include <errno.h>

#include <signal.h>
#include <syslog.h>

#include <pcap.h>

#include "amesis.h"
#include "dumper.h"

/* Global vars */

/* Am i a daemon ? */
int bg = 0;
unsigned char sf = 0; 

int
log_info(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	if (bg)
		vsyslog(LOG_INFO, fmt, ap);
	else
		vprintf(fmt, ap);
	va_end(ap);
	return 0;
}

int
log_warn(const char *fmt,...)
{
	va_list ap;
	va_start(ap, fmt);
	if (bg)
		vsyslog(LOG_ALERT, fmt, ap);
	else
		vfprintf(stderr, fmt, ap);
	va_end(ap);
	return 0;
}

int
log_debug(const char *fmt,...)
{
	va_list ap;

	if (bg)
		return -1;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);

	return 0;
}

int
usage(void)
{
	log_info("> usage\n");

	return 0;
}

void
sigfunc(int sig)
{
	log_warn("signal %d recieved: exiting.\n", sig);
	sf = 1;
}


int
handle_packets(pifs tab_if, const unsigned char *pack, struct pcap_pkthdr h)
{
	/*
	 * printf("packet size off wire: %d\n", h.caplen);
	 * printf("packet size: %d\n", h.len);
	 */

	dump_packet(tab_if, pack, h);

	return 0;
}

int
link_len(int type)
{
	switch(type) {
	case DLT_IEEE802_11:
		return 32;
	case DLT_EN10MB:
		return 14;
	case DLT_SLIP:
		return 16;
	case DLT_NULL:
	case DLT_LOOP:
		return 4;
	case DLT_PPP:
	case DLT_C_HDLC:
	case DLT_PPP_SERIAL:
		return 4;
	case DLT_PPP_ETHER:
		return 8;
	case DLT_IEEE802:
		return 22;
	case DLT_PRISM_HEADER:	
		return 144+30;
	case DLT_ATM_RFC1483:
		return 8;
	case DLT_RAW:
		return 0;
	case DLT_ATM_CLIP:
		return 8;
	case DLT_LINUX_SLL:
		return 16;
	case DLT_LTALK:
		return 0;
	}
	return -1;
}

int
main(int argc, char **argv)
{
	int ch, ret;
	unsigned int i, j, ok;
	char *p;
	char *tmpdev;

	struct sigaction sig;

	struct timeval ts;
	int snaplen = DEF_SNAPLEN;

	/* General bpf rule */
	char *g_bpf_rule = NULL;

	/* All interfaces */
	pifs tab_if[MAX_IF], tab_if_copy[MAX_IF];
	unsigned int num_if = 0, num_if_ok = 0;
	int max = 0;
	fd_set fds, fds_save;

	/* pcap bpf stuff */
	struct bpf_program fcode;
	struct pcap_stat pcapstat;

	/* pcap error buffer */
	char errbuf[PCAP_ERRBUF_SIZE];

	const unsigned char *packet;
	struct pcap_pkthdr h;

	while ((ch = getopt (argc, argv, "c:df:g:hi:l:r:s:")) != -1) {
		p = argv[optind - 1];
		switch (ch) {
		case 'c':	/* configuration file */
			break;
		case 'd':	/* foreground mode */
			bg = 1;
			break;
		case 'f':
		case 'i':	/* add a interface to listen */
			if (num_if < MAX_IF) {
				tab_if[num_if] = malloc(sizeof(ifs));
				if (tab_if[num_if] == NULL) {
					log_warn("malloc: %s\n", 
					         strerror(errno));
				}

				if (ch == 'f')
					tab_if[num_if]->mode = MODE_READ;
				else
					tab_if[num_if]->mode = MODE_LIVE;

				tab_if[num_if]->if_name = strdup(p);
				tab_if[num_if]->bpf_rule = NULL;
				tab_if[num_if]->pfd = NULL;
				tab_if[num_if]->log_name = NULL;
				tab_if[num_if]->log_desc = NULL;
				num_if ++;
			}
			break;
		case 'l':
			if(num_if != 0) {
				tab_if[num_if-1]->log_name = strdup(p);
			} else {
				log_warn("can't use -l without -i\n");
				exit(EXIT_FAILURE);
			}
			break;
		case 'r':	/* add a rule to last -i interface */
			if(num_if != 0) {
				tab_if[num_if-1]->bpf_rule = strdup(p);
			} else {
				log_warn("can't use -r without -i\n");
				exit(EXIT_FAILURE);
			}
			break;
		case 's':
			snaplen = strtol(p, NULL, 10);
			break;
		case 'g':	/* default rule for interfaces */ 
			g_bpf_rule = strdup(p);
			break;
		case 'h':
		default:
			usage();
			exit(EXIT_SUCCESS);
			break;
		}
	}

	/* Signal handlers */
	sig.sa_handler = sigfunc;
	sig.sa_flags = 0;

	sigaction(SIGINT, &sig, NULL);

	printf("snaplen is at %d\n", snaplen);

	/* Interface init. */

	if (num_if == 0) {
		*tab_if = malloc(sizeof(ifs));
		if (*tab_if == NULL) {
			log_warn("malloc: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		tmpdev = pcap_lookupdev(errbuf);
		if (tmpdev == NULL) {
			log_warn("lookupdev: %s\n", errbuf);
			free(*tab_if);
			exit(EXIT_FAILURE);
		}
		tab_if[0]->if_name = strdup(tmpdev);
		tab_if[0]->pfd = NULL;	
		tab_if[0]->bpf_rule = NULL;
		tab_if[0]->mode = MODE_LIVE;
		num_if ++;
	}

	for (i = 0; i < num_if; i ++) {
		if (tab_if[i]->mode == MODE_READ) {
			tab_if[i]->pfd = pcap_open_offline(tab_if[i]->if_name,
			                                   errbuf);
		} 

		if (tab_if[i]->mode == MODE_LIVE) {
			tab_if[i]->pfd = pcap_open_live(tab_if[i]->if_name,
		                                        snaplen, 1, 1,
		                                        errbuf);
		}

		if (tab_if[i]->pfd == NULL) {
			log_warn("pcap_open(%s): %s\n", tab_if[i]->if_name,
			                                errbuf);
			free(tab_if[i]->if_name);
			tab_if[i]->if_name = NULL;
			if (tab_if[i]->bpf_rule != NULL) {
				free(tab_if[i]->bpf_rule);
				tab_if[i]->bpf_rule = NULL;
			}
			free(tab_if[i]);
			tab_if[i] = NULL;
		} else {
			if (tab_if[i]->log_name != NULL) {
				tab_if[i]->log_desc = 
					pcap_dump_open(tab_if[i]->pfd,
					               tab_if[i]->log_name);

				if(tab_if[i]->log_desc == NULL) {
					log_warn("pcap_dump_open(%s/%s): %s\n",
						tab_if[i]->if_name,
						tab_if[i]->log_name,
						pcap_geterr(tab_if[i]->pfd));
					free(tab_if[i]->log_name);
					tab_if[i]->log_name = NULL;
				}
			}
			num_if_ok ++;
			tab_if[i]->link_type = pcap_datalink(tab_if[i]->pfd);
			tab_if[i]->linkh_len = link_len(tab_if[i]->link_type);
			
			ret = pcap_lookupnet(tab_if[i]->if_name,
			                    &tab_if[i]->localnet,
			                    &tab_if[i]->netmask, 
			                     errbuf);

			if (tab_if[i]->mode == MODE_LIVE)
				if (max < pcap_fileno(tab_if[i]->pfd))
					max = pcap_fileno(tab_if[i]->pfd);

			if (tab_if[i]->mode == MODE_READ) {
				if (max < fileno(pcap_file(tab_if[i]->pfd)))
					max= fileno(pcap_file(tab_if[i]->pfd));
			}

			if (ret < 0) {
				log_info("lookupnet: %s\n", errbuf);
			} 

			if(tab_if[i]->bpf_rule == NULL && g_bpf_rule != NULL) {
				tab_if[i]->bpf_rule = strdup(g_bpf_rule);
			} 

			if(tab_if[i]->bpf_rule != NULL) {

				ret = pcap_compile(tab_if[i]->pfd, 
				                  &fcode,
				                   tab_if[i]->bpf_rule,
				                   1, tab_if[i]->localnet);
				if(ret < 0) {
					log_info("error on: %s\n",
					         tab_if[i]->bpf_rule);
					free(tab_if[i]->bpf_rule);
					tab_if[i]->bpf_rule = NULL;
					continue;
				}

				ret = pcap_setfilter(tab_if[i]->pfd, &fcode);
				if(ret < 0) {
					log_info("error on: %s\n",
					         tab_if[i]->bpf_rule);
				}
			}
		}
	}

	for (i = 0, j = 0; i < num_if; i ++) {
		if (tab_if[i] != NULL) { 
			tab_if_copy[j++] = tab_if[i];
		}
	}

	memcpy(tab_if, tab_if_copy, sizeof(tab_if_copy));

	if (num_if_ok == 0) {
		log_warn("No interface, exiting.\n");
		exit(EXIT_FAILURE);
	}

	if (bg != 0) {
		daemon(0, 0);
	} else {
		setvbuf(stdout, NULL,  _IONBF, 0);
	}

	/* Main part */
	ts.tv_sec = 3;
	ts.tv_usec = 0;

	FD_ZERO(&fds_save);
	for (i = 0; i < num_if_ok; i ++) {
		if (tab_if[i] != NULL) {
			if(tab_if[i]->mode == MODE_LIVE)
				FD_SET(pcap_fileno(tab_if[i]->pfd), &fds_save);
			if(tab_if[i]->mode == MODE_READ) {
				if (pcap_file(tab_if[i]->pfd) != NULL) {
					ret= fileno(pcap_file(tab_if[i]->pfd));
					FD_SET(ret , &fds_save);
				}
			}
		}
	}

	fds = fds_save;
	while((ret = select(max + 1, &fds, NULL, NULL, &ts)) != -1 && sf == 0){
		for(i = 0; i < num_if_ok; i ++) {

			if (tab_if[i]->mode == MODE_DEAD) {
				continue ;
			}

			if (tab_if[i]->mode == MODE_LIVE) {
				ret = pcap_fileno(tab_if[i]->pfd);
				if (!FD_ISSET(ret, &fds))
					continue ;
			}
			if (tab_if[i]->mode == MODE_READ) {
				ret = fileno(pcap_file(tab_if[i]->pfd));
				if (!FD_ISSET(ret, &fds))
					continue ;

			}

			packet = pcap_next(tab_if[i]->pfd, &h);

			if (packet == NULL) {
				/* Don't close yet, i need it for stats. */
				/* pcap_close (tab_if[i]->pfd); */
				if (tab_if[i]->mode == MODE_READ) {
					FD_CLR(ret, &fds_save);
					tab_if[i]->mode = MODE_DEAD;
				} else {
					ok = 1;
				}
				
				/* ok, may be there is no other thing */
				ok = 0;
				for (j = 0; j < num_if_ok && ok == 0; j ++) {
					if (tab_if[j]->mode == MODE_LIVE)
						ok = 1;
				}

				if (ok == 0) {
					log_info("No more stream.\n");
					goto fin_select;
				}
			} else {
				/* We need to log, as asked. */
				if (tab_if[i]->log_desc != NULL) {
					pcap_dump((u_char*)tab_if[i]->log_desc, 
					          &h, packet);
				}
				handle_packets(tab_if[i], packet, h);
			}
		}

		ts.tv_sec = 3;
		fds = fds_save;
	}

fin_select:

	for (i = 0; i < num_if_ok; i ++) {
		/* pcap_stats(pcap_t *p, struct pcap_stat *ps) */
		if (pcap_stats(tab_if[i]->pfd, &pcapstat) == 0) {
			printf("Packets received: %d\n", pcapstat.ps_recv);
			printf("Packets dropped:  %d\n", pcapstat.ps_drop);
		}
		if (tab_if[i]->log_desc != NULL) {
			pcap_dump_close(tab_if[i]->log_desc);
		}
		pcap_close(tab_if[i]->pfd);
		free(tab_if[i]->if_name);
		if (tab_if[i]->bpf_rule != NULL) {
			free(tab_if[i]->bpf_rule);
		}
	}

	return 0;
}
